/*	$Id: viamDisplay.c 2010/12/06 00:42:27 tho $ */

/* Copyright (c) 2007-2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Sun Dec 30 2007
 */

#include <portLib.h>
#include <time.h>
#include <err.h>
#include <pthread.h>

#include <viam/viamlib.h>

#include "server/viamHeader.h"
#include "shared.h"

#ifdef VIAM_DISPLAY

/* --- local data ---------------------------------------------------- */

static CvFont		cvfont;		/**< performance tag font */

static pthread_t	dspthread;	/**< image display thread */
static pthread_cond_t	dspsignal;	/**< update display when signalled */

static pthread_mutex_t	dsplistlock;	/**< protect display list */
static viam_bank_t *	dsplist;	/**< display list */
static int		dsplistlength;	/**< active items in display list */
static int		dsplistsize;	/**< display list length */

static void *		viamDisplayThread(void *dummy);


/* --- viamDisplayInit ----------------------------------------------- */

/** Initialize display
 *
 */

STATUS
viamDisplayInit(int *report)
{
  /* performance tags font */
  cvInitFont(&cvfont, CV_FONT_HERSHEY_PLAIN, 1., 1., 0., 1, CV_AA);

  /* display list */
  if (pthread_cond_init(&dspsignal, NULL)) goto error;
  if (pthread_mutex_init(&dsplistlock, NULL)) goto error;
  dsplist = NULL;
  dsplistlength = dsplistsize = 0;

  /* display thread */
  if (pthread_create(&dspthread, NULL, viamDisplayThread, NULL)) goto error;

  return OK;
error:
  *report = S_viam_NO_DISPLAY_THREAD;
  return ERROR;
}


/* --- viamDisplaySignal --------------------------------------------- */

/** Signal display thread
 *
 */

void
viamDisplaySignal(const viam_bank_t bank)
{
  int i;

  if (pthread_mutex_lock(&dsplistlock)) return;

  for(i=0; i<dsplistlength; i++)
    if (dsplist[i] == bank) goto done;

  if (dsplistlength >= dsplistsize) {
    viam_bank_t *list = realloc(dsplist, (dsplistlength+1)*sizeof(*dsplist));
    if (!list) goto done;

    dsplist = list;
    dsplistsize = dsplistlength+1;
  }
  dsplist[dsplistlength++] = bank;
  pthread_cond_signal(&dspsignal);

done:
  pthread_mutex_unlock(&dsplistlock);
}


/* --- viamDisplayThread --------------------------------------------- */

/** Display thread
 *
 */

void *
viamDisplayThread(void *dummy)
{
  viam_bank_t *list;	/**< display list */
  int length;		/**< active items in display list */
  int size;		/**< display list length */
  char *dbuffer;	/**< temporary image buffer */
  size_t dsize;		/**< size of dbuffer */
  IplImage image;
  int i, j, k;

  list = NULL;
  length = size = 0;
  dbuffer = NULL;
  dsize = 0;

  while(1) {
    if (pthread_mutex_lock(&dsplistlock)) return NULL;
    if (dsplistlength == 0)
      do
	pthread_cond_wait(&dspsignal, &dsplistlock);
      while (dsplistlength == 0);

    if (dsplistlength > size) {
      viam_bank_t *l = realloc(list, (dsplistlength)*sizeof(*list));
      if (!l) {
	pthread_mutex_unlock(&dsplistlock);
	continue;
      }
      list = l;
      size = dsplistlength;
    }
    memcpy(list, dsplist, dsplistlength*sizeof(*list));
    length = dsplistlength;

    dsplistlength = 0;
    pthread_mutex_unlock(&dsplistlock);

    for(i=0; i<length; i++) {
      viam_bankdata_t *bd = viam_bank_usrdata(list[i]);

      for(j=0; j<viam_bank_nimages(list[i]); j++) {
	viam_imagedata_t *idata = viamIData(list[i], j);

	/* copy image as we may modify it  */
	image = *viamPosterCurrentImage(vh, bd->pdata, j);
	if (dsize < image.imageSize) {
	  dbuffer = realloc(dbuffer, image.imageSize);
	  if (!dbuffer) continue;
	  dsize = image.imageSize;
	}
	memcpy(dbuffer, image.imageData, image.imageSize);
	cvSetData(&image, dbuffer, image.widthStep);

	/* rescale 32F images between 0-1 */
	if (image.depth == IPL_DEPTH_32F) {
	  double min = 0., max = 0.;
	  for(k=0; k<image.width*image.height; k++) {
	    if (((float *)image.imageData)[k] < min)
	      min = ((float *)image.imageData)[k];
	    if (((float *)image.imageData)[k] > max)
	      max = ((float *)image.imageData)[k];
	  }
	  if (min == max) max = min + 1.;
	  for(k=0; k<image.width*image.height; k++) {
	    ((float *)image.imageData)[k] =
	      (((float *)image.imageData)[k] - min) / (max - min);
	  }
	}

	/* performance tag */
	if (idata->vtags == VIAM_ON) {
	  CvSize isize, tsize;
	  char vtag[256];
	  int baseline;

	  snprintf(vtag, sizeof(vtag), "%s %4.1f fps",
		   ctime(&idata->tacq.tv_sec), idata->fps);

	  isize = cvGetSize(&image);
	  cvfont.hscale = (double)isize.width/idata->wwidth;
	  cvfont.vscale = (double)isize.height/idata->wheight;
	  cvfont.thickness = cvfont.hscale;

	  cvGetTextSize(vtag, &cvfont, &tsize, &baseline);
	  if (tsize.width > idata->wwidth) {
	    cvfont.hscale *= (double)isize.width/tsize.width;
	    cvfont.thickness = cvfont.hscale;
	    cvGetTextSize(vtag, &cvfont, &tsize, &baseline);
	  }
	  cvPutText(&image, vtag,
		    cvPoint(isize.width - tsize.width - 4,
			    isize.height - tsize.height - 4),
		    &cvfont, cvScalarAll(0));
	  cvPutText(&image, vtag,
		    cvPoint(isize.width - tsize.width - 5,
			    isize.height - tsize.height - 5),
		    &cvfont, cvScalarAll(255));
	}

	/* display */
	if (idata->window) cvShowImage(idata->window, &image);
      }
    }

    /* Handle display events, for at most 20ms (50Hz is enough for display) */
    cvWaitKey(20);
  }

  return NULL;
}

#endif /* VIAM_DISPLAY */
